import pandas as pd
from scipy.constants import Avogadro


def mass_to_nr_molecules(mass, mm):
    """Calculating the number of molecules form mass and molarmass.

    Mass, Molar mass need to be in the same unit, either both g and g/mol
    or kg and kg/mol.

    Parameters
    ---------------------------
    mass : float
        mass of all compounds.

    mm : float
        molar mass of compounds.

    Returns
    -----------------
    nr_molecules : float
        number of molecules
    """
    nr_molecules = mass / mm * Avogadro
    return nr_molecules


def nr_molecules_to_mass(nr_molecules, mm):
    """Calculates the mass from the number of molecules and molar mass.

    Parameters
    ---------------
    nr_molecules : int
        Number of molecules

    mm : float
        Molar mass [g/mol]

    Returns
    ---------------------
    mass : float
        mass in grams
    """
    mass = mm * nr_molecules / Avogadro
    return mass


def unitconv_sfc_conc_bck(data, x=2):
    """Converting: ugSOx/m3 to  ugS/ m3.

    Parameters
    ------------------
    data: ndarray
        Contains the data in units of ugSoX/m3.

    x: int
        The number of oxygen atoms, O in you desired SOx compound.

    Returns
    ------------
    data : ndarray
        in units of ugS/ m3.

    Notes
    -----------
    micro grams to kilos is 10**6
    """

    mmO = 15.9999  # molar mass oxygen
    mmS = 32.065  # molar mass sulphur
    mm_compound = (mmS + x * mmO) * 10**3  # *10**3 gives molar mass in micrograms

    nr_molecules = mass_to_nr_molecules(data, mm_compound)
    weight_s = nr_molecules_to_mass(nr_molecules, mmS * 10**3)  # weigth in ug
    return weight_s


def unitconv_sfc_conc(data, nr_of_O=2):
    """Unitconverting:  ugS/m3 to ugSOx/m3

    Parameters
    ------------------
    data : array_like
        Contains the data in units of ugS/m3.

    nr_of_O: int
        The number of O's in you desired SOx compound.

    Returns
    ------------
    data : ndarray
        data in units of ug SOx/m3

    """

    mm_s = 32.065 * 10**6  # in units of ug/mol
    mm_o = nr_of_O * 15.9999 * 10**6  ## in units of ug/mol
    nr_molecules = mass_to_nr_molecules(data, mm_s)  # 32.065*10**6) [ug/mol]
    added_weight_oksygen = nr_molecules_to_mass(nr_molecules, mm_o)  # ug
    # added weights in micrograms
    mass = data + added_weight_oksygen  # in micrograms
    return mass


def unitconv_wet_depo_bck(data, time, ts_type="monthly"):
    """The unitconversion kg SO4 m-2 s-1 to kgS/ha.

    Removing the weight of oxygen.

    Parameters
    ------------------
    data: ndarray
        Sulphur data you wish to convert.

    time : pd.Seires[numpy.datetime64]
         Array of datetime64 timesteps.

    ts_type: str
       The timeseries type. Default monthly.

    Returns
    ------------------
    data : ndarray
       Sulphur data in units of ugSOx m-3 s-1.

    """
    # kg SO4 m-2 s-1 to kg S/ha
    mm_so4 = 0.001 * 32.065 + 0.001 * 15.999 * 4  # kg/mol
    mm_s = 32.065 * 0.001  # kg/mol

    days_in_month = time.dt.daysinmonth
    monthly_to_sec = days_in_month * 24 * 60 * 60  # Seconds in each
    nr_molecules = data * Avogadro / mm_so4  # [1]
    mass_S = nr_molecules * mm_s / Avogadro  # mass in kg
    # Mulitply by seconds in one month
    mass_pr_ha = mass_S * monthly_to_sec * 10000
    return mass_pr_ha


def unitconv_wet_depo(data, time, ts_type="monthly"):
    """Unitconversion kg S/ha to kg SOx m-2 s-1.

    Adding mass of oksygen.

    Parameters
    ------------------
    data: ndarray
        data in unit kg S/ha = kg S/(1000 m2)

    time : pd.Seires[numpy.datetime64]
        Array of datetime64 timesteps.

    ts_type : str
        The timeseries type. Default "monthly".

    Returns
    ------------------
    data : ndarray
        data in units of ugSOx/m3

    """
    mmSO4 = 0.001 * 32.065 + 0.001 * 15.999 * 4  # in kg/mol
    mm_s = 32.065 * 0.001  # kg/mol
    # print('uses new updated version, in that case fix conversion')
    nr_molecules = data * Avogadro / mm_s  # [1]
    mass_SO4 = nr_molecules * mmSO4 / Avogadro  # mass in kl
    days_in_month = time.dt.daysinmonth
    monthly_to_sec = days_in_month * 24 * 60 * 60
    # print('includes new changes')
    mass_pr_square_m_pr_sek = mass_SO4 / (10000 * monthly_to_sec)
    return mass_pr_square_m_pr_sek


def unitconv_wet_depo_from_emep(data, time, ts_type="monthly"):
    """Unitconversion mgS m-2 to kg SO4 m-2 s-1.

    Milligram to kilos is 10-6.

    Adding mass of oksygen.

    Parameters
    ------------------
    data: ndarray
        data in unit mg S m-2.

    time : pd.Seires[numpy.datetime64]
        Array of datetime64 timesteps.

    ts_type : str
        The timeseries type. Default "monthly".

    Returns
    ------------------
    data : ndarray
        data in units of ugSOx/m3

    """

    if not isinstance(time, pd.Series):
        time = pd.Series(time)

    mm_so4 = 0.001 * 32.065 + 0.001 * 15.999 * 4  # in kg/mol
    mm_s = 0.001 * 32.065  # kg/mol
    data_in_kilos = data * 10 ** (-9)
    nr_molecules = data_in_kilos * Avogadro / mm_s  # [1]
    mass_SO4 = nr_molecules * mm_so4 / Avogadro  # mass in kg
    days_in_month = time.dt.daysinmonth
    monthly_to_sec = days_in_month * 24 * 60 * 60
    mass_pr_square_m_pr_sek = mass_SO4 * 10000 / monthly_to_sec
    return mass_pr_square_m_pr_sek



from __future__ import annotations

import pytest

from pyaerocom.variable import Variable


@pytest.mark.parametrize(
    "var_name,init,kwargs",
    [
        (None, True, {}),
        ("od550aer", None, {"bla": 42}),
        ("od550aer", None, {"map_vmin": 0, "map_vmax": 1}),
        ("concpm10", None, {}),
        ("concpm103d", None, {}),
        ("concpm103D", None, {}),
    ],
)
def test_Variable(var_name: str | None, init: bool, kwargs: dict):
    var = Variable(var_name, init, **kwargs)
    for key, val in kwargs.items():
        assert getattr(var, key) == val


@pytest.mark.parametrize(
    "var_name,cfg,error",
    [
        pytest.param(
            None,
            "bla",
            "invalid input for cfg, need config parser got <class 'str'>",
            id="not a config parser",
        ),
        pytest.param(
            "bla_blub",
            None,
            "invalid variable name bla_blub. Must not contain underscore",
            id="variable name with underscore",
        ),
    ],
)
def test_Variable_error(var_name: str | None, cfg, error: str):
    with pytest.raises(ValueError) as e:
        Variable(var_name, True, cfg)
    assert str(e.value) == error


@pytest.mark.parametrize(
    "var_name,var_name_aerocom",
    [
        ("od550aer3D", "od550aer"),
        ("od550aer", "od550aer"),
        ("od550du", "od550dust"),
        ("od550csaer", "od550aer"),
        ("latitude", "lat"),
        ("abs550oc", "abs550oa"),
        ("sconcss", "concss"),
    ],
)
def test_Variable_var_name_aerocom(var_name: str, var_name_aerocom: str):
    assert Variable(var_name).var_name_aerocom == var_name_aerocom


def test_Variable_alias_var():
    assert "od550csaer" == Variable("od550aer")


def test_Variable_alias_families():
    var = Variable("sconcso4")

    assert var.var_name_input == "sconcso4"
    assert var.var_name == "sconcso4"
    assert var.var_name_aerocom == "concso4"
    assert var.units == "ug m-3"


@pytest.mark.parametrize(
    "var_name,result",
    [
        ("od550aer", False),
        ("emiso4", True),
        ("depso4", False),
        ("pr", False),
        ("prmm", False),
        ("dryso4", False),
        ("wetso4", False),
    ],
)
def test_Variable_is_emission(var_name: str, result: bool):
    assert Variable(var_name).is_emission == result


@pytest.mark.parametrize(
    "var_name,result",
    [
        ("od550aer", False),
        ("emiso4", False),
        ("depso4", True),
        ("pr", False),
        ("prmm", False),
        ("dryso4", True),
        ("wetso4", True),
    ],
)
def test_Variable_is_deposition(var_name: str, result: bool):
    assert Variable(var_name).is_deposition == result


@pytest.mark.parametrize(
    "var_name,result",
    [
        ("od550aer", False),
        ("emiso4", True),
        ("depso4", True),
        ("pr", True),
        ("prmm", True),
        ("dryso4", True),
        ("wetso4", True),
    ],
)
def test_Variable_is_rate(var_name: str, result: bool):
    assert Variable(var_name).is_rate == result


def test_Variable___str__():
    var = Variable("od550aer")
    s = str(var)
    assert s.startswith("\nPyaerocom Variable")
    assert "var_name: od550aer" in s



"""
Metadata keys from GHOST that only depend on station dimension

THESE SHOULD BE ALWAYS THE SAME ACCORDING TO INFORMATION INFORMATION
PROVIDED BY BSC
"""

from pyaerocom.io.ghost.meta_keys import ghost_meta_keys

desired_keys = [
    "EDGAR_v4.3.2_annual_average_BC_emissions",
    "EDGAR_v4.3.2_annual_average_CO_emissions",
    "EDGAR_v4.3.2_annual_average_NH3_emissions",
    "EDGAR_v4.3.2_annual_average_NMVOC_emissions",
    "EDGAR_v4.3.2_annual_average_NOx_emissions",
    "EDGAR_v4.3.2_annual_average_OC_emissions",
    "EDGAR_v4.3.2_annual_average_PM10_emissions",
    "EDGAR_v4.3.2_annual_average_SO2_emissions",
    "EDGAR_v4.3.2_annual_average_biogenic_PM2.5_emissions",
    "EDGAR_v4.3.2_annual_average_fossilfuel_PM2.5_emissions",
    "ESDAC_Iwahashi_landform_classification",
    "ESDAC_Meybeck_landform_classification",
    "ESDAC_modal_Iwahashi_landform_classification_25km",
    "ESDAC_modal_Iwahashi_landform_classification_5km",
    "ESDAC_modal_Meybeck_landform_classification_25km",
    "ESDAC_modal_Meybeck_landform_classification_5km",
    "ETOPO1_altitude",
    "ETOPO1_max_altitude_difference_5km",
    "GHOST_version",
    "GPW_average_population_density_25km",
    "GPW_average_population_density_5km",
    "GPW_max_population_density_25km",
    "GPW_max_population_density_5km",
    "GPW_population_density",
    "GSFC_coastline_proximity",
    "Joly-Peuch_classification_code",
    "Koppen-Geiger_classification",
    "Koppen-Geiger_modal_classification_25km",
    "Koppen-Geiger_modal_classification_5km",
    "MODIS_MCD12C1_v6_IGBP_land_use",
    "MODIS_MCD12C1_v6_LAI",
    "MODIS_MCD12C1_v6_UMD_land_use",
    "MODIS_MCD12C1_v6_modal_IGBP_land_use_25km",
    "MODIS_MCD12C1_v6_modal_IGBP_land_use_5km",
    "MODIS_MCD12C1_v6_modal_LAI_25km",
    "MODIS_MCD12C1_v6_modal_LAI_5km",
    "MODIS_MCD12C1_v6_modal_UMD_land_use_25km",
    "MODIS_MCD12C1_v6_modal_UMD_land_use_5km",
    "NOAA-DMSP-OLS_v4_average_nighttime_stable_lights_25km",
    "NOAA-DMSP-OLS_v4_average_nighttime_stable_lights_5km",
    "NOAA-DMSP-OLS_v4_max_nighttime_stable_lights_25km",
    "NOAA-DMSP-OLS_v4_max_nighttime_stable_lights_5km",
    "NOAA-DMSP-OLS_v4_nighttime_stable_lights",
    "OMI_level3_column_annual_average_NO2",
    "OMI_level3_column_cloud_screened_annual_average_NO2",
    "OMI_level3_tropospheric_column_annual_average_NO2",
    "OMI_level3_tropospheric_column_cloud_screened_annual_average_NO2",
    "UMBC_anthrome_classification",
    "UMBC_modal_anthrome_classification_25km",
    "UMBC_modal_anthrome_classification_5km",
    "WMO_region",
    "WWF_TEOW_biogeographical_realm",
    "WWF_TEOW_biome",
    "WWF_TEOW_terrestrial_ecoregion",
    "administrative_country_division_1",
    "administrative_country_division_2",
    "altitude",
    "area_classification",
    "associated_networks",
    "city",
    "climatology",
    "contact_email_address",
    "contact_institution",
    "contact_name",
    "country",
    "daily_passing_vehicles",
    "data_level",
    "daytime_traffic_speed",
    "distance_to_building",
    "distance_to_junction",
    "distance_to_kerb",
    "distance_to_source",
    "land_use",
    "latitude",
    "longitude",
    "main_emission_source",
    "measurement_altitude",
    "measurement_methodology",
    "measurement_scale",
    "measuring_instrument_calibration_scale",
    "measuring_instrument_documented_absorption_cross_section",
    "measuring_instrument_documented_accuracy",
    "measuring_instrument_documented_flow_rate",
    "measuring_instrument_documented_lower_limit_of_detection",
    "measuring_instrument_documented_measurement_resolution",
    "measuring_instrument_documented_precision",
    "measuring_instrument_documented_span_drift",
    "measuring_instrument_documented_uncertainty",
    "measuring_instrument_documented_upper_limit_of_detection",
    "measuring_instrument_documented_zero_drift",
    "measuring_instrument_documented_zonal_drift",
    "measuring_instrument_further_details",
    "measuring_instrument_inlet_information",
    "measuring_instrument_manual_name",
    "measuring_instrument_name",
    "measuring_instrument_process_details",
    "measuring_instrument_reported_absorption_cross_section",
    "measuring_instrument_reported_accuracy",
    "measuring_instrument_reported_flow_rate",
    "measuring_instrument_reported_lower_limit_of_detection",
    "measuring_instrument_reported_measurement_resolution",
    "measuring_instrument_reported_precision",
    "measuring_instrument_reported_span_drift",
    "measuring_instrument_reported_uncertainty",
    "measuring_instrument_reported_units",
    "measuring_instrument_reported_upper_limit_of_detection",
    "measuring_instrument_reported_zero_drift",
    "measuring_instrument_reported_zonal_drift",
    "measuring_instrument_sampling_type",
    "network",
    "network_maintenance_details",
    "network_miscellaneous_details",
    "network_provided_volume_standard_pressure",
    "network_provided_volume_standard_temperature",
    "network_qa_details",
    "network_sampling_details",
    "network_uncertainty_details",
    "population",
    "primary_sampling_further_details",
    "primary_sampling_instrument_documented_flow_rate",
    "primary_sampling_instrument_manual_name",
    "primary_sampling_instrument_name",
    "primary_sampling_instrument_reported_flow_rate",
    "primary_sampling_process_details",
    "primary_sampling_type",
    "principal_investigator_email_address",
    "principal_investigator_institution",
    "principal_investigator_name",
    "process_warnings",
    "representative_radius",
    "sample_preparation_further_details",
    "sample_preparation_process_details",
    "sample_preparation_techniques",
    "sample_preparation_types",
    "sampling_height",
    "station_classification",
    "station_name",
    "station_reference",
    "station_timezone",
    "street_type",
    "street_width",
    "terrain",
]


def test_ghost_meta_keys():
    assert sorted(ghost_meta_keys()) == sorted(desired_keys)



import logging
import warnings
from ast import literal_eval
from configparser import ConfigParser

import numpy as np

from pyaerocom import var_groups
from pyaerocom._lowlevel_helpers import dict_to_str, list_to_shortstr
from pyaerocom.exceptions import VariableDefinitionError
from pyaerocom.mathutils import make_binlist
from pyaerocom.obs_io import OBS_WAVELENGTH_TOL_NM

#: helper vor checking if variable name contains str 3d or 3D
from pyaerocom.variable_helpers import (
    _check_alias_family,
    _read_alias_ini,
    get_aliases,
    parse_aliases_ini,
    parse_variables_ini,
)
from pyaerocom.varnameinfo import VarNameInfo

logger = logging.getLogger(__name__)


def literal_eval_list(val: str):
    return list(literal_eval(val))


def str2list(val: str) -> list[str]:
    return [x.strip() for x in val.split(",")]


def str2bool(val: str) -> bool:
    return val.lower() in {"true", "1", "t", "yes"}


class Variable:
    """Interface that specifies default settings for a variable

    See `variables.ini <https://github.com/metno/pyaerocom/blob/master/
    pyaerocom/data/variables.ini>`__ file for an overview of currently available
    default variables.

    Parameters
    ----------
    var_name : str
        string ID of variable (see file variables.ini for valid IDs)
    init : bool
        if True, input variable name is attempted to be read from config file
    cfg : ConfigParser
        open config parser that holds the information in config file available
        (i.e. :func:`ConfigParser.read` has been called with config file as
        input)
    **kwargs
        any valid class attribute (e.g. map_vmin, map_vmax, ...)

    Attributes
    ----------
    var_name : str
        input variable name
    var_name_aerocom : str
        AEROCOM variable name (see e.g. `AEROCOM protocol
        <http://aerocom.met.no/protocol_table.html>`__ for a list of
        available variables)
    is_3d : bool
        flag that indicates if variable is 3D
    is_dry : bool
        flag that is set based on filename that indicates if variable data
        corresponds to dry conditions.
    units : str
        unit of variable (None if no unit)
    default_vert_code : str, optional
        default vertical code to be loaded (i.e. Column, ModelLevel, Surface).
        Only relevant during reading and in case conflicts occur (e.g.
        abs550aer, 2010, Column and Surface files)
    aliases : list
        list of alternative names for this variable
    minimum : float
        lower limit of allowed value range
    upper_limit : float
        upper limit of allowed value range
    obs_wavelength_tol_nm : float    literal_eval_list = lambda val: list(literal_eval(val))

        wavelength tolerance (+/-) for reading of obsdata. Default is 10, i.e.
        if this variable is defined at 550 nm and obsdata contains measured
        values of this quantity within interval of 540 - 560, then these data
        is used
    scat_xlim : float
        x-range for scatter plot
    scat_ylim : float
        y-range for scatter plot
    scat_loglog : bool
        scatter plot on loglog scale
    scat_scale_factor : float
        scale factor for scatter plot
    map_cmap : str
        name of default colormap (matplotlib) of this variable.
    map_vmin : float
        data value corresponding to lower end of colormap in map plots of this
        quantity
    map_vmax : float
        data value corresponding to upper end of colormap in map plots of this
        quantity
    map_c_under : str
        color used for values below :attr:`map_vmin` in map plots of this
        quantity
    map_c_over : str
        color used for values exceeding :attr:`map_vmax` in map plots of this
        quantity
    map_cbar_levels : :obj:`list`, optional
        levels of colorbar
    map_cbar_ticks : :obj:`list`, optional
        colorbar ticks
    """

    _TYPE_CONV = {
        "wavelength_nm": float,
        "minimum": float,
        "maximum": float,
        "dimensions": str2list,
        "obs_wavelength_tol_nm": float,
        "scat_xlim": literal_eval_list,
        "scat_ylim": literal_eval_list,
        "scat_loglog": str2bool,
        "scat_scale_factor": float,
        "dry_rh_max": float,
        "map_cmap": str,
        "map_vmin": float,
        "map_vmax": float,
        "map_cbar_levels": literal_eval_list,
        "map_cbar_ticks": literal_eval_list,
        "_is_rate": bool,
    }

    # maybe used in config
    ALT_NAMES = {"unit": "units"}

    plot_info_keys = [
        "scat_xlim",
        "scat_ylim",
        "scat_loglog",
        "scat_scale_factor",
        "map_vmin",
        "map_vmax",
        "map_cmap",
        "map_c_under",
        "map_c_over",
        "map_cbar_levels",
        "map_cbar_ticks",
    ]
    VMIN_DEFAULT = -np.inf
    VMAX_DEFAULT = np.inf

    @staticmethod
    def _check_input_var_name(var_name):
        if "3d" in var_name:
            var_name = var_name.replace("3d", "")
        elif "3D" in var_name:
            var_name = var_name.replace("3D", "")
        elif "_" in var_name:
            raise ValueError(f"invalid variable name {var_name}. Must not contain underscore")
        return var_name

    def __init__(self, var_name=None, init=True, cfg=None, **kwargs):
        if var_name is None:
            var_name = "od550aer"
        elif not isinstance(var_name, str):
            raise ValueError(
                f"Invalid input for variable name, need str type, got {type(var_name)}"
            )
        # save orig. input for whatever reason
        self._var_name_input = var_name

        self.var_name = self._check_input_var_name(var_name)
        self._var_name_aerocom = None

        self.standard_name = None
        # Assume variables that have no unit specified in variables.ini are
        # unitless.
        self.units = "1"
        self.default_vert_code = None

        self.wavelength_nm = None
        self.dry_rh_max = 40
        self.dimensions = None
        self.minimum = self.VMIN_DEFAULT
        self.maximum = self.VMAX_DEFAULT

        self.description = None
        self.comments_and_purpose = None

        # wavelength tolerance in nm
        self.obs_wavelength_tol_nm = None

        self.scat_xlim = None
        self.scat_ylim = None
        self.scat_loglog = None
        self.scat_scale_factor = 1.0

        # settings for map plotting
        self.map_cmap = "coolwarm"
        self.map_vmin = None
        self.map_vmax = None
        self.map_c_under = None
        self.map_c_over = "r"
        self.map_cbar_levels = None
        self.map_cbar_ticks = None

        self._is_rate = False

        # imports default information and, on top, variable information (if
        # applicable)
        if init:
            self.parse_from_ini(self.var_name, cfg=cfg)

        self.update(**kwargs)
        if self.obs_wavelength_tol_nm is None:
            self.obs_wavelength_tol_nm = OBS_WAVELENGTH_TOL_NM

    @property
    def var_name_aerocom(self):
        """AeroCom variable name of the input variable"""
        vna = self._var_name_aerocom
        return self.var_name if vna is None else vna

    @property
    def var_name_input(self):
        """Input variable"""
        return self._var_name_input

    @property
    def is_3d(self):
        """True if str '3d' is contained in :attr:`var_name_input`"""
        return True if "3d" in self.var_name_input.lower() else False

    @property
    def is_wavelength_dependent(self):
        """Indicates whether this variable is wavelength dependent"""
        return True if self.wavelength_nm is not None else False

    @property
    def is_at_dry_conditions(self):
        """Indicate whether variable denotes dry conditions"""
        var_name = self.var_name_aerocom
        if var_name.startswith("dry"):  # dry deposition
            return False
        return True if "dry" in var_name else False

    @property
    def is_deposition(self):
        """
        Indicates whether input variables is a deposition rate

        Note
        ----
        This funtion only identifies wet and dry deposition based on the variable
        names, there might be other variables that are deposition variables but
        cannot be identified by this function.

        Parameters
        ----------
        var_name : str
            Name of variable to be checked

        Returns
        -------
        bool
            If True, then variable name denotes a deposition variables

        """
        var_name = self.var_name_aerocom
        if var_name.startswith(var_groups.drydep_startswith):
            return True
        elif var_name.startswith(var_groups.wetdep_startswith):
            return True
        elif var_name.startswith(var_groups.totdep_startswith):
            return True
        elif var_name in var_groups.dep_add_vars:
            return True
        return False

    @property
    def is_emission(self):
        """
        Indicates whether input variables is an emission rate

        Note
        ----
        This funtion only identifies wet and dry deposition based on the variable
        names, there might be other variables that are deposition variables but
        cannot be identified by this function.

        Parameters
        ----------
        var_name : str
            Name of variable to be checked

        Returns
        -------
        bool
            If True, then variable name denotes a deposition variables

        """
        var_name = self.var_name_aerocom
        if var_name.startswith(var_groups.emi_startswith):
            return True
        elif var_name in var_groups.emi_add_vars:
            return True
        return False

    @property
    def is_rate(self):
        """Indicates whether variable name is a rate

        Rates include e.g. deposition or emission rate variables but also
        precipitation

        Returns
        -------
        bool
            True if variable is rate, else False
        """
        if self.is_emission:
            return True
        elif self.is_deposition:
            return True
        elif self._is_rate:
            return True
        return False

    @property
    def is_alias(self):
        return True if self.var_name != self.var_name_aerocom else False

    @property
    def unit(self):
        """Unit of variable (old name, deprecated)"""
        warnings.warn(
            "Attr. name unit in Variable class is deprecated. Please use units instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.units

    @property
    def plot_info(self):
        """Dictionary containing plot information"""
        d = {}
        for k in self.plot_info_keys:
            d[k] = self[k]
        return d

    def update(self, **kwargs):
        for key, val in kwargs.items():
            self[key] = val

    @property
    def has_unit(self):
        """Boolean specifying whether variable has unit"""
        return True if self.units not in (1, None) else False

    @property
    def lower_limit(self):
        """Old attribute name for :attr:`minimum` (following HTAP2 defs)"""
        warnings.warn("Old name for attribute minimum", DeprecationWarning, stacklevel=2)
        return self.minimum

    @property
    def upper_limit(self):
        """Old attribute name for :attr:`maximum` (following HTAP2 defs)"""
        warnings.warn("Old name for attribute maximum", DeprecationWarning, stacklevel=2)
        return self.maximum

    @property
    def unit_str(self):
        """string representation of unit"""
        if self.units is None:
            return ""
        else:
            return f"[{self.units}]"

    @staticmethod
    def read_config():
        return parse_variables_ini()

    @property
    def var_name_info(self):
        return VarNameInfo(self.var_name)

    @property
    def aliases(self):
        """Alias variable names that are frequently found or used

        Returns
        -------
        list
            list containing valid aliases
        """
        return get_aliases(self.var_name)

    @property
    def long_name(self):
        """Wrapper for :attr:`description`"""
        return self.description

    def keys(self):
        return list(self.__dict__)

    @staticmethod
    def _check_aliases(var_name):
        ap = parse_aliases_ini()
        aliases = _read_alias_ini(ap)
        if var_name in aliases:
            return aliases[var_name]
        return _check_alias_family(var_name, ap)

    def get_default_vert_code(self):
        """Get default vertical code for variable name"""
        if self.default_vert_code is not None:
            return self.default_vert_code
        try:
            return VarNameInfo(self.var_name_aerocom).get_default_vert_code()
        except ValueError:
            logger.warning(
                f"default_vert_code not set for {self.var_name_aerocom} and "
                f"could also not be inferred"
            )
            return None

    def get_cmap(self):
        """
        Get cmap str for var

        Returns
        -------
        str

        """
        return self.map_cmap

    def _cmap_bins_from_vmin_vmax(self):
        """
        Calculate cmap discretisation bins from :attr:`vmin` and :attr:`vmax`

        Sets value of :attr:`map_cbar_levels`

        Raises
        ------
        AttributeError
             if :attr:`vmin` and :attr:`vmax` are not defined

        """
        if self.minimum == self.VMIN_DEFAULT or self.maximum == self.VMAX_DEFAULT:
            raise AttributeError(
                f"need minimum and maximum to be specified "
                f"for variable {self.var_name} in "
                f"order to retrieve cmap_bins"
            )
        self.map_cbar_levels = make_binlist(self.minimum, self.maximum)

    def get_cmap_bins(self, infer_if_missing=True):
        """
        Get cmap discretisation bins

        Parameters
        ----------
        infer_if_missing : bool
            if True and :attr:`map_cbar_levels` is not defined, try to infer
            using :func:`_cmap_bins_from_vmin_vmax`.

        Raises
        ------
        AttributeError
             if unavailable

        Returns
        -------
        list
            levels

        """
        if self.map_cbar_levels is None:
            if infer_if_missing:
                self._cmap_bins_from_vmin_vmax()
            else:
                raise AttributeError(
                    f"map_cbar_levels is not defined for variable {self.var_name}"
                )
        return self.map_cbar_levels

    def parse_from_ini(self, var_name=None, cfg=None):
        """Import information about default region

        Parameters
        ----------
        var_name : str
            variable name
        var_name_alt : str
            alternative variable name that is used if variable name is not
            available
        cfg : ConfigParser
            open config parser object

        Returns
        -------
        bool
            True, if default could be loaded, False if not
        """
        if cfg is None:
            cfg = self.read_config()
        elif not isinstance(cfg, ConfigParser):
            raise ValueError(f"invalid input for cfg, need config parser got {type(cfg)}")
        if var_name not in cfg:
            try:
                var_name = self._check_aliases(var_name)
            except VariableDefinitionError:
                logger.info(f"Unknown input variable {var_name}")
                return
            self._var_name_aerocom = var_name

        var_info = cfg[var_name]
        # this variable should import settings from another variable
        if "use" in var_info:
            use = var_info["use"]
            if use not in cfg:
                raise VariableDefinitionError(
                    f"Input variable {var_name} depends on {use} "
                    f"which is not available in variables.ini."
                )
            self.parse_from_ini(use, cfg)

        for key, val in var_info.items():
            if key in self.ALT_NAMES:
                key = self.ALT_NAMES[key]
            self._add(key, val)

    def _add(self, key, val):
        if key in self._TYPE_CONV:
            try:
                val = self._TYPE_CONV[key](val)
            except Exception:
                pass
        elif key == "units" and val == "None":
            val = "1"
        if val == "None":
            val = None
        self[key] = val

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __getitem__(self, key):
        return self.__dict__[key]

    def __repr__(self):
        return f"{self.var_name}\nstandard_name: {self.standard_name}; Unit: {self.units}"

    def __eq__(self, other):
        if isinstance(other, str):
            other = Variable(other)
        elif not isinstance(other, Variable):
            raise TypeError("Can only compare with str or other Variable instance")
        return True if other.var_name_aerocom == self.var_name_aerocom else False

    def __str__(self):
        head = f"Pyaerocom {type(self).__name__}"
        s = f"\n{head}\n{len(head)*'-'}"

        plot_s = "\nPlotting settings\n......................"

        for k, v in self.__dict__.items():
            if k in self.plot_info_keys:
                if v is None:
                    continue
                if isinstance(v, dict):
                    plot_s += f"\n{k} (dict)"
                    plot_s += dict_to_str(v, indent=3, ignore_null=True)
                elif isinstance(v, list):
                    plot_s += f"\n{k} (list, {len(v)} items)"
                    plot_s += list_to_shortstr(v)
                else:
                    plot_s += f"\n{k}: {v}"
            else:
                if isinstance(v, dict):
                    s += f"\n{k} (dict)"
                    s += dict_to_str(v, indent=3, ignore_null=True)
                elif isinstance(v, list):
                    s += f"\n{k} (list, {len(v)} items)"
                    s += list_to_shortstr(v)
                else:
                    s += f"\n{k}: {v}"

        s += plot_s
        return s



import numpy as np
import pandas as pd
import pytest
from scipy.constants import Avogadro

from pyaerocom._warnings import ignore_warnings
from pyaerocom.io.helpers_units import (
    mass_to_nr_molecules,
    nr_molecules_to_mass,
    unitconv_sfc_conc,
    unitconv_sfc_conc_bck,
    unitconv_wet_depo_bck,
    unitconv_wet_depo_from_emep,
)


@pytest.mark.parametrize(
    "mass,nm, result",
    [
        (0, 1, 0),
        (1, 1, Avogadro),
    ],
)
def test_mass_to_nr_molecules(mass, nm, result):
    assert result == mass_to_nr_molecules(mass, nm)


@pytest.mark.parametrize(
    "nr_molecules,mm, result",
    [
        (0, 1, 0),
        (1, Avogadro, 1),
    ],
)
def test_nr_molecules_to_mass(nr_molecules, mm, result):
    assert result == nr_molecules_to_mass(nr_molecules, mm)


@pytest.fixture()
def dummy_data():
    return np.ones(10)


def test_unitconv_sfc_conc_bck(dummy_data):
    result = unitconv_sfc_conc_bck(dummy_data)
    assert len(result) == len(dummy_data)
    assert np.all(result == pytest.approx(0.50050886, 1e-4))


def test_unitconv_sfc_conc(dummy_data):
    result = unitconv_sfc_conc(dummy_data)
    assert len(result) == len(dummy_data)
    assert np.all(result == pytest.approx(1.99796663, 1e-4))


@ignore_warnings(
    FutureWarning,
    "'M' is deprecated and will be removed in a future version, please use 'ME' instead.",
)
# TODO: The above warning is ignored because the old-dependency CI tests don't currently
# support ME. Once we bump dependencies so we can change over, this should be removed.
def test_unitconv_wet_depo_bck(dummy_data):
    time = pd.Series(pd.date_range(start="2000-01-01", periods=len(dummy_data), freq="M"))
    result = unitconv_wet_depo_bck(dummy_data, time)
    assert len(result) == len(
        dummy_data
    )  # sufficent to check length b/c wet depo will change month-to-month


@ignore_warnings(
    FutureWarning,
    "'M' is deprecated and will be removed in a future version, please use 'ME' instead.",
)
# TODO: The above warning is ignored because the old-dependency CI tests don't currently
# support ME. Once we bump dependencies so we can change over, this should be removed.
def test_unitconv_wet_depo_from_emep(dummy_data):
    time = pd.Series(pd.date_range(start="2000-01-01", periods=len(dummy_data), freq="M"))
    result = unitconv_wet_depo_from_emep(dummy_data, time)
    assert len(result) == len(
        dummy_data
    )  # sufficent to check length b/c wet depo will change month-to-month


@ignore_warnings(
    FutureWarning,
    "'M' is deprecated and will be removed in a future version, please use 'ME' instead.",
)
# TODO: The above warning is ignored because the old-dependency CI tests don't currently
# support ME. Once we bump dependencies so we can change over, this should be removed.
def test_unitconv_wet_depo_from_emep_time_not_pandas_series(dummy_data):
    time = pd.date_range(start="2000-01-01", periods=len(dummy_data), freq="M")
    result = unitconv_wet_depo_from_emep(dummy_data, time)
    assert len(result) == len(dummy_data)



import logging
import os
import sys
from collections.abc import Callable
from functools import cached_property
from pathlib import Path
from typing import Literal

import pandas as pd
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ValidationError,
    field_validator,
    model_validator,
)

from pyaerocom import const
from pyaerocom._lowlevel_helpers import LayerLimits, RegridResDeg
from pyaerocom.config import ALL_REGION_NAME
from pyaerocom.helpers import start_stop
from pyaerocom.io.pyaro.pyaro_config import PyaroConfig

logger = logging.getLogger(__name__)


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class ColocationSetup(BaseModel):
    """
    Setup class for high-level model / obs co-location.

    An instance of this setup class can be used to run a colocation analysis
    between a model and an observation network and will create a number of
    :class:`pya.ColocatedData` instances, which can be saved automatically
    as NetCDF files.

    Apart from co-location, this class also handles reading of the input data
    for co-location. Supported co-location options are:

    1. gridded vs. ungridded data
    For instance 3D model data (instance of :class:`GriddedData`) with lat,
    lon and time dimension that is co-located with station based observations
    which are represented in pyaerocom through :class:`UngriddedData` objects.
    The co-location function used is
    :func:`pyaerocom.colocation.colocated_gridded_ungridded`. For this type of
    co-location, the output co-located data object will be 3-dimensional,
    with dimensions `data_source` (index 0: obs, index 1: model), `time` and
    `station_name`.

    2. gridded vs. gridded data
    For instance 3D model data that is co-located with 3D satellite data
    (both instances of :class:`GriddedData`), both objects with lat,
    lon and time dimensions. The co-location function used
    is :func:`pyaerocom.colocation.colocated_gridded_gridded`.
    For this type of co-location, the output co-located data object will be
    4-dimensional,  with dimensions `data_source` (index 0: obs, index 1:
    model), `time` and `latitude` and `longitude`.



    Attributes
    ----------
    model_id : str
        ID of model to be used.

    pyaro_config: PyaroConfig
        In the case Pyaro is used, a config must be provided. In that case obs_id(see below)
        is ignored and only the config is used.
    obs_id : str
        ID of observation network to be used.
    obs_vars : tuple[str, ...]
        Variables to be analysed (need to be available in input obs dataset).
        Variables that are not available in the model data output will be
        skipped. Alternatively, model variables to be used for a given obs
        variable can also be specified via attributes :attr:`model_use_vars`
        and :attr:`model_add_vars`.
    ts_type : str
        String specifying colocation output frequency.
    start
        Start time of colocation. Input can be integer denoting the year or
        anything that can be converted  into :class:`pandas.Timestamp` using
        :func:`pyaerocom.helpers.to_pandas_timestamp`. If None, than the first
        available date in the model data is used.
    stop
        stop time of colocation. int or anything that can be converted into
        :class:`pandas.Timestamp` using
        :func:`pyaerocom.helpers.to_pandas_timestamp` or None. If None and if
        ``start`` is on resolution of year (e.g. ``start=2010``) then ``stop``
        will be automatically set to the end of that year. Else, it will be
        set to the last available timestamp in the model data.
    filter_name : str
        name of filter to be applied. If None, no filter is used
        (to be precise, if None, then
         :attr:`pyaerocom.const.DEFAULT_REG_FILTER` is used which should
         default to `ALL-wMOUNTAINS`, that is, no filtering).
    basedir_coldata : str | Path
        Base directory for storing of colocated data files.
    save_coldata : bool
        if True, colocated data objects are saved as NetCDF file.
    obs_name : str, optional
        if provided, this string will be used in colocated data filename to
        specify obsnetwork, else obs_id will be used.
    obs_data_dir : str, optional
        location of obs data. If None, attempt to infer obs location based on
        obs ID.
    obs_use_climatology : bool
        BETA if True, pyaerocom default climatology is computed from observation
        stations (so far only possible for unrgidded / gridded colocation).
    obs_vert_type : str
        AeroCom vertical code encoded in the model filenames (only AeroCom 3
        and later). Specifies which model file should be read in case there are
        multiple options (e.g. surface level data can be read from a
        *Surface*.nc file as well as from a *ModelLevel*.nc file). If input is
        string (e.g. 'Surface'), then the corresponding vertical type code is
        used for reading of all variables that are colocated (i.e. that are
        specified in :attr:`obs_vars`).
    obs_ts_type_read : str or dict, optional
        may be specified to explicitly define the reading frequency of the
        observation data (so far, this does only apply to gridded obsdata such
        as satellites), either as str (same for all obs variables) or variable
        specific as dict. For ungridded reading, the frequency may be specified
        via :attr:`obs_id`, where applicable (e.g. AeronetSunV3Lev2.daily).
        Not to be confused with :attr:`ts_type`, which specifies the
        frequency used for colocation. Can be specified variable specific in
        form of dictionary.
    obs_filters : dict
        filters applied to the observational dataset before co-location.
        In case of gridded / gridded, these are filters that can be passed to
        :func:`pyaerocom.io.ReadGridded.read_var`, for instance, `flex_ts_type`,
        or `constraints`. In case the obsdata is ungridded (gridded / ungridded
        co-locations) these are filters that are handled through keyword
        `filter_post` in :func:`pyaerocom.io.ReadUngridded.read`. These filters
        are applied to the :class:`UngriddedData` objects after reading and
        caching the data, so changing them, will not invalidate the latest
        cache of the :class:`UngriddedData`.
    read_opts_ungridded : dict, optional
        dictionary that specifies reading constraints for ungridded reading,
        and are passed as `**kwargs` to :func:`pyaerocom.io.ReadUngridded.read`.
        Note that - other than for `obs_filters` these filters are applied
        during the reading of the :class:`UngriddedData` objects and specifying
        them will deactivate caching.
    model_name : str, optional
        if provided, this string will be used in colocated data filename to
        specify model, else obs_id will be used.
    model_data_dir : str, optional
        Location of model data. If None, attempt to infer model location based
        on model ID.
    model_read_opts : dict, optional
        options for model reading (passed as keyword args to
        :func:`pyaerocom.io.ReadUngridded.read`).
    model_use_vars : dict, optional
        dictionary that specifies mapping of model variables. Keys are
        observation variables, values are the corresponding model variables
        (e.g. model_use_vars=dict(od550aer='od550csaer')). Example: your
        observation has var *od550aer* but your model model uses a different
        variable name for that variable, say *od550*. Then, you can specify
        this via `model_use_vars = {'od550aer' : 'od550'}`. NOTE: in this case,
        a model variable *od550aer* will be ignored, even if it exists
        (cf :attr:`model_add_vars`).
    model_rename_vars : dict, optional
        rename certain model variables **after** co-location, before storing
        the associated :class:`ColocatedData` object on disk. Keys are model
        variables, values are new names
        (e.g. `model_rename_vars={'od550aer':'MyAOD'}`).
        Note: this does not impact which variables are read from the model.
    model_add_vars : dict, optional
        additional model variables to be processed for one obs variable. E.g.
        `model_add_vars={'od550aer': ['od550so4', 'od550gt1aer']}` would
        co-locate both model SO4 AOD (od550so4) and model coarse mode AOD
        (od550gt1aer) with total AOD (od550aer) from obs (in addition to
        od550aer vs od550aer if applicable).
    model_to_stp : bool
        ALPHA (please do not use): convert model data values to STP conditions
        after co-location. Note: this only works for very particular settings
        at the moment and needs revision, as it relies on access to
        meteorological data.
    model_ts_type_read : str or dict, optional
        may be specified to explicitly define the reading frequency of the
        model data, either as str (same for all obs variables) or variable
        specific as dict. Not to be confused with :attr:`ts_type`, which
        specifies the output frequency of the co-located data.
    model_read_aux : dict, optional
        may be used to specify additional computation methods of variables from
        models. Keys are variables to be computed, values are dictionaries with
        keys `vars_required` (list of required variables for computation of var
        and `fun` (method that takes list of read data objects and computes
        and returns var).
    model_use_climatology : bool
        if True, attempt to use climatological model data field. Note: this
        only works if model data is in AeroCom conventions (climatological
        fields are indicated with 9999 as year in the filename) and if this is
        active, only single year analysis are supported (i.e. provide int to
        :attr:`start` to specify the year and leave :attr:`stop` empty).
    model_kwargs: dict
        Key word arguments to be given to the model reader class's read_var and init function
    gridded_reader_id : dict
        BETA: dictionary specifying which gridded reader is supposed to be used
        for model (and gridded obs) reading. Note: this is a workaround
        solution and will likely be removed in the future when the gridded
        reading API is more harmonised
        (see https://github.com/metno/pyaerocom/issues/174).
    flex_ts_type : bool
        Bboolean specifying whether reading frequency of gridded data is
        allowed to be flexible. This includes all gridded data, whether it is
        model or gridded observation (e.g. satellites). Defaults to True.
    min_num_obs : dict or int, optional
        time resampling constraints applied, defaults to None, in which case
        no constraints are applied. For instance, say your input is in daily
        resolution and you want output in monthly and you want to make sure to
        have roughly 50% daily coverage for the monthly averages. Then you may
        specify `min_num_obs=15` which will ensure that at least 15 daily
        averages are available to compute a monthly average. However, you may
        also define a hierarchical scheme that first goes from daily to
        weekly and then from weekly to monthly, via a dict. E.g.
        `min_num_obs=dict(monthly=dict(weekly=4), weekly=dict(daily=3))` would
        ensure that each week has at least 3 daily values, as well as that each
        month has at least 4 weekly values.
    resample_how : str or dict, optional
        string specifying how data should be aggregated when resampling in time.
        Default is "mean". Can also be a nested dictionary, e.g.
        `resample_how={'conco3': 'daily': {'hourly' : 'max'}}` would use the
        maximum value to aggregate from hourly to daily for variable conco3,
        rather than the mean.
    obs_remove_outliers : bool
        if True, outliers are removed from obs data before colocation,
        else not. Default is False.
        Custom outlier ranges for each variable can be specified via
        :attr:`obs_outlier_ranges`, and for all other variables, the pyaerocom
        default outlier ranges are used. The latter are specified in
        `variables.ini` file via `minimum` and `maximum` attributes and can
        also be accessed through :attr:`pyaerocom.variable.Variable.minimum`
        and :attr:`pyaerocom.variable.Variable.maximum`, respectively.
    model_remove_outliers : bool
        if True, outliers are removed from model data (normally this should be
        set to False, as the models are supposed to be assessed, including
        outlier cases). Default is False.
        Custom outlier ranges for each variable can be specified via
        :attr:`model_outlier_ranges`, and for all other variables, the pyaerocom
        default outlier ranges are used. The latter are specified in
        `variables.ini` file via `minimum` and `maximum` attributes and can
        also be accessed through :attr:`pyaerocom.variable.Variable.minimum`
        and :attr:`pyaerocom.variable.Variable.maximum`, respectively.
    obs_outlier_ranges : dict, optional
        dictionary specifying outlier ranges for individual obs variables.
        (e.g. dict(od550aer = [-0.05, 10], ang4487aer=[0,4])). Only relevant
        if :attr:`obs_remove_outliers` is True.
    model_outlier_ranges : dict, optional
        like :attr:`obs_outlier_ranges` but for model variables. Only relevant
        if :attr:`model_remove_outliers` is True.
    zeros_to_nan : bool
        If True, zero's in output co-located data object will be converted to
        NaN. Default is False.
    harmonise_units : bool
        if True, units are attempted to be harmonised during co-location
        (note: raises Exception if True and in case units cannot be harmonised).
    regrid_res_deg : float or dict, optional
        regrid resolution in degrees. If specified, the input gridded data
        objects will be regridded in lon / lat dimension to the input
        resolution (if input is float, both lat and lon are regridded to that
        resolution, if input is dict, use keys `lat_res_deg` and `lon_res_deg`
        to specify regrid resolutions, respectively). Default is None.
    colocate_time : bool
        if True and if obs and model sampling frequency (e.g. daily) are higher
        than output colocation frequency (e.g. monthly), then the datasets are
        first colocated in time (e.g. on a daily basis), before the monthly
        averages are calculated. Default is False.
    reanalyse_existing : bool
        if True, always redo co-location, even if there is already an existing
        co-located NetCDF file (under the output location specified by
        :attr:`basedir_coldata` ) for the given variable combination to be
        co-located. If False and output already exists, then co-location is
        skipped for the associated variable. This flag is also used for contour-plots.
        Default is True.
    raise_exceptions : bool
        if True, Exceptions that may occur for individual variables to be
        processed, are raised, else the analysis is skipped for such cases.
    keep_data : bool
        if True, then all colocated data objects computed when running
        :func:`run` will be stored in :attr:`data`. Defaults to True.
    add_meta : dict
        additional metadata that is supposed to be added to each output
        :class:`ColocatedData` object.
    main_freq:
        Main output frequency for AeroVal (some of the AeroVal processing
        steps are only done for this resolution, since they would create too
        much output otherwise, such as statistics timeseries or scatter plot in
        "Overall Evaluation" tab on AeroVal).
        Note that this frequency needs to be included in next setting "freqs".
    freqs:
        Frequencies for which statistical parameters are computed
    """

    ##########################
    # Pydantic ConfigDict
    ##########################
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        allow="extra",
        protected_namespaces=(),
        # TODO
        # frozen=True,  # make immutable
        # validate_assignment=True,
    )

    #########################
    # Init Input
    #########################

    model_id: str | None
    obs_id: str | None
    obs_vars: tuple[str, ...] | str

    @field_validator("obs_vars")
    @classmethod
    def validate_obs_vars(cls, v):
        if isinstance(v, str):
            return (v,)
        return v

    ts_type: str
    start: pd.Timestamp | int | str | None
    stop: pd.Timestamp | int | str | None

    @field_validator("start", "stop")
    @classmethod
    def validate_start_stop(cls, v):
        if isinstance(v, int):
            return v
        if isinstance(v, str):
            return pd.Timestamp(v)

    pyaro_config: PyaroConfig | None = None

    ###############################
    # Attributes with defaults
    ###############################

    #: Dictionary specifying alternative vertical types that may be used to
    #: read model data. E.g. consider the variable is  ec550aer,
    #: obs_vert_type='Surface' and obs_vert_type_alt=dict(Surface='ModelLevel').
    #: Now, if a model that is used for the analysis does not contain a data
    #: file for ec550aer at the surface ('*ec550aer*Surface*.nc'), then, the
    #: colocation routine will look for '*ec550aer*ModelLevel*.nc' and if this
    #: exists, it will load it and extract the surface level.
    OBS_VERT_TYPES_ALT: dict[str, str] = {"Surface": "ModelLevel", "2D": "2D"}

    #: do not raise Exception if invalid item is attempted to be assigned
    #: (Overwritten from base class)
    CRASH_ON_INVALID: bool = False

    FORBIDDEN_KEYS: list[str] = [
        "var_outlier_ranges",  # deprecated since v0.12.0
        "var_ref_outlier_ranges",  # deprecated since v0.12.0
        "remove_outliers",  # deprecated since v0.12.0
    ]

    # crashes if input filter name is invalid
    filter_name: str = f"{ALL_REGION_NAME}-wMOUNTAINS"

    basedir_coldata: str | Path = Field(default=const.COLOCATEDDATADIR, validate_default=True)

    @field_validator("basedir_coldata")
    @classmethod
    def validate_basedirs(cls, v):
        if not os.path.exists(v):
            tmp = Path(v) if isinstance(v, str) else v
            tmp.mkdir(parents=True, exist_ok=True)
        return v

    save_coldata: bool = False

    # Options related to obs reading and processing
    obs_name: str | None = None
    obs_data_dir: Path | str | None = None

    obs_use_climatology: bool = False

    obs_cache_only: bool = False  # only relevant if obs is ungridded
    obs_vert_type: str | None = None
    obs_ts_type_read: str | dict | None = None
    obs_filters: dict = {}
    colocation_layer_limits: tuple[LayerLimits, ...] | None = None
    profile_layer_limits: tuple[LayerLimits, ...] | None = None
    read_opts_ungridded: dict | None = {}

    # Attributes related to model data
    model_name: str | None = None
    model_data_dir: Path | str | None = None

    model_read_opts: dict | None = {}

    model_use_vars: dict[str, str] | None = {}
    model_rename_vars: dict[str, str] | None = {}
    model_add_vars: dict[str, tuple[str, ...]] | None = {}
    model_to_stp: bool = False

    model_ts_type_read: str | dict | None = None
    model_read_aux: (
        dict[str, dict[Literal["vars_required", "fun"], list[str] | Callable]] | None
    ) = {}
    model_use_climatology: bool = False

    gridded_reader_id: dict[str, str] = {"model": "ReadGridded", "obs": "ReadGridded"}

    flex_ts_type: bool = True

    # Options related to time resampling
    min_num_obs: dict | int | None = None
    resample_how: str | dict | None = "mean"

    # Options related to outlier removal
    obs_remove_outliers: bool = False
    model_remove_outliers: bool = False

    # Custom outlier ranges for model and obs
    obs_outlier_ranges: dict[str, tuple[float, float]] | None = {}
    model_outlier_ranges: dict[str, tuple[float, float]] | None = {}
    zeros_to_nan: bool = False
    harmonise_units: bool = False
    regrid_res_deg: float | RegridResDeg | None = None
    colocate_time: bool = False
    reanalyse_existing: bool = True
    raise_exceptions: bool = False
    keep_data: bool = True
    add_meta: dict | None = {}

    model_kwargs: dict = {}

    main_freq: str = "monthly"
    freqs: list[str] = ["monthly", "yearly"]

    @field_validator("model_kwargs")
    @classmethod
    def validate_kwargs(cls, v):
        forbidden = [
            "vert_which",
        ]  # Forbidden key names which are not found in colocation_setup.model_field, or has another name there
        for key in v:
            if key in list(cls.model_fields.keys()) + forbidden:
                raise ValueError(f"Key {key} not allowed in model_kwargs")
        return v

    # Override __init__ to allow for positional arguments
    def __init__(
        self,
        model_id: str | None = None,
        pyaro_config: PyaroConfig | None = None,
        obs_id: str | None = None,
        obs_vars: tuple[str, ...] | None = (),
        ts_type: str = "monthly",
        start: pd.Timestamp | int | None = None,
        stop: pd.Timestamp | int | None = None,
        basedir_coldata: str = const.COLOCATEDDATADIR,
        save_coldata: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(
            model_id=model_id,
            pyaro_config=pyaro_config,
            obs_id=obs_id,
            obs_vars=obs_vars,
            ts_type=ts_type,
            start=start,
            stop=stop,
            basedir_coldata=basedir_coldata,
            save_coldata=save_coldata,
            **kwargs,
        )

    # Model validator for forbidden keys
    @model_validator(mode="after")
    def validate_no_forbidden_keys(self):
        for key in self.FORBIDDEN_KEYS:
            if key in self.model_fields:
                raise ValidationError
        return self

    @cached_property
    def basedir_logfiles(self):
        p = Path(self.basedir_coldata) / "logfiles"
        if not p.exists():
            p.mkdir(parents=True, exist_ok=True)
        return str(p)

    @model_validator(mode="after")
    def validate_pyaro_config(self):
        if self.pyaro_config is None:
            return self
        if self.pyaro_config.name != self.obs_id:
            logger.info(
                f"Data ID in Pyaro config {self.pyaro_config.name} does not match obs_id {self.obs_id}. Setting Pyaro config to None!"
            )
            self.pyaro_config = None
        if self.pyaro_config is not None:
            if isinstance(self.pyaro_config, dict):
                logger.info("pyaro config was given as dict. Will try to convert to PyaroConfig")
                self.pyaro_config = PyaroConfig(**self.pyaro_config)
            if self.pyaro_config.name != self.obs_id:
                logger.info(
                    f"Data ID in Pyaro config {self.pyaro_config.name} does not match obs_id {self.obs_id}. Setting Obs ID to match Pyaro Config!"
                )
                self.obs_id = self.pyaro_config.name
            if self.obs_id is None:
                self.obs_id = self.pyaro_config.name
        return self

    def add_glob_meta(self, **kwargs):
        """
        Add global metadata to :attr:`add_meta`

        Parameters
        ----------
        kwargs
            metadata to be added

        Returns
        -------
        None

        """
        self.add_meta.update(**kwargs)

    def _period_from_start_stop(self) -> str:
        start, stop = start_stop(self.start, self.stop, stop_sub_sec=False)
        y0, y1 = start.year, stop.year
        assert y0 <= y1
        if y0 == y1:
            return str(y0)
        else:
            return f"{y0}-{y1}"

    def update(self, data: dict) -> Self:
        # provide an update() method analogous to MutableMapping's one

        # validate values in data
        update = self.model_dump()
        update.update(data)
        self.model_validate(update)
        # assign values from data
        for k, v in data.items():
            logger.debug(f"updating value of '{k}' from '{getattr(self, k, None)}' to '{v}'")
            setattr(self, k, v)
        return self



import pytest

from pyaerocom.io import ReadEEAAQEREP_V2
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData
from tests.conftest import TEST_RTOL, lustre_avail

# Subset data id used for testing.
DATA_ID = "EEA_AQeRep.v2.Subset"


@pytest.fixture(scope="module")
def reader():
    return ReadEEAAQEREP_V2(DATA_ID)


def test_get_file_list(reader):
    # at this point that is the base directory without recursive search
    # so this returns only Revision.txt and metadata.csv
    # don't be too restrictive since we might have additional files in the subdirectory
    assert len(reader.get_file_list()) >= 2


@lustre_avail
def test_read(reader):
    # special station codes to test
    # not sure if these are really needed
    # station ids to test
    station_id = {}
    # respective mean for a station; index has to be the same as station_id
    station_means = {}
    # list of tested variables will be extended
    var_name = "concpm10"
    # station #3 has in time unordered data
    # station #4 has just one time step
    station_id[var_name] = [
        "AT10002",
        "AT52000",
        "AT90TAB",
        "AT4S418",
    ]
    station_means[var_name] = [
        17.128,
        15.1,
        26.968,
        77.098,
    ]

    var_names_to_test = station_id.keys()
    for var_name in var_names_to_test:
        data = None
        data = reader.read(vars_to_retrieve=[var_name])
        assert isinstance(data, UngriddedData)

        for stat_idx, statid in enumerate(station_id[var_name]):
            stat_data = data[statid]
            # It makes no sense to test this for every station
            if stat_idx == 1:
                assert isinstance(stat_data, StationData)

            assert stat_data[var_name].mean() == pytest.approx(
                station_means[var_name][stat_idx], TEST_RTOL
            )
